/* MC.RC - Icon and version info resource file

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
#include <windows.h>
#include <winver.h>
#include "config.h"

/////////////////////////////////////////////////////////////////////////////
// Version stamp for this .EXE
VS_VERSION_INFO     VERSIONINFO
  FILEVERSION       1,0,0,1
  PRODUCTVERSION    3,0,0,1
  FILEFLAGSMASK     VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS         VS_FF_DEBUG|VS_FF_PRIVATEBUILD|VS_FF_PRERELEASE
#else
  FILEFLAGS         0 // final version
#endif
  FILEOS            VOS_NT
  FILETYPE          VFT_APP
  FILESUBTYPE       0   // not used
BEGIN
	BLOCK "StringFileInfo"
	BEGIN			   
		BLOCK "040904E4" // Lang=US English, CharSet=Windows Multilingual
		BEGIN 
			VALUE "CompanyName",     "The Freeware Fundation\0"
			VALUE "FileDescription", "GNU Midnight Commander\0"
			VALUE "FileVersion",     VERSION
			VALUE "InternalName",    "MC\0"
			VALUE "LegalCopyright",  "(c) The Freeware Fundation\0"
			VALUE "LegalTrademarks", "see GNU General Public License\0"
			VALUE "OriginalFilename","MC.EXE\0"
			VALUE "ProductName",     "GNU Midnight Commander\0"
			VALUE "ProductVersion",  VERSION
		END
	END
	BLOCK "VarFileInfo" 
	BEGIN 
		VALUE "Translation", 0x409, 1252
			// English language (0x409) and the Windows ANSI codepage (1252)
	END
END

/////////////////////////////////////////////////////////////////////////////
// ICON
0 ICON "mc.ico"
